# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unohelper
import traceback
from operator import itemgetter
from itertools import groupby
from collections import defaultdict
from itertools import groupby
import locale
from com.sun.star.awt import XDialogEventHandler
from com.sun.star.text.ControlCharacter import PARAGRAPH_BREAK
from com.sun.star.container import NoSuchElementException
from com.sun.star.beans import PropertyValue, UnknownPropertyException

# some tweaks are needed
import sys
locale.setlocale(locale.LC_ALL, '')
if sys.version_info < (3, ):
    reload(sys)
    sys.setdefaultencoding(sys.getfilesystemencoding())


class IndexCreator(unohelper.Base, XDialogEventHandler):
    SEP = '|'   # separator for key entries
    PATTERN = '{} \t {}'
    CONTENT_SUFFIX = "_index_content"
    HEADER_SUFFIX = "_index_header"
    STYLES = {1: "User Index 1", 2: "User Index 2", 3: "User Index 3", 4: "User Index 4", 5: "User Index 5",
              6: "User Index 6", 7: "User Index 7", 8: "User Index 8", 9: "User Index 9", 10: "User Index 10"}
    LOCALES = dict(TITLE = {'it': "Indici analitici", 'fr': 'Index lexicaux'},
                   LBL_CHOICE = {'it': "Schegli l'indice:", 'fr': "Choisir l'index :"},
                   LBL_INSERT = {'it': "INSERISCI", 'fr': 'INSÉRER'},
                   LBL_UPDATE = {'it': "AGGIORNA", 'fr': 'METTRE À JOUR'},
                   LBL_DELETE = {'it': "CANCELLA", 'fr': 'SUPPRIMER'},
                   LBL_CANCEL = {'it': "ANNULA", 'fr': 'ANNULER'},
                   LBL_ALL = {'it': "Applica a ~tutti", 'fr': 'Appliquer à tous'},
                   MSG_INSERTED = {'it': 'indice "{}" inserito', 'fr': 'index {} inséré'},
                   MSG_CANCELLED = {'it': 'indice "{}" cancellato', 'fr': 'index {} supprimé'},
                   MSG_UPDATE = {'it': 'indice "{}" aggiornato', 'fr': 'index {} mis à jour'})

    def __init__(self, ctx, doc, loc='fr'):
        if not loc in self.availablelocales():
            loc = 'fr'
        self.loadlocale(loc)
        self.ctx = ctx
        self.smgr = ctx.ServiceManager
        self.doc = doc
        self.viewcursor = doc.CurrentController.ViewCursor
        try:
            self.currentpos = self.viewcursor.Start
            self.vctext = self.viewcursor.Text
        except UnknownPropertyException:
            return
        self.undo = self.doc.UndoManager
        self.text = doc.Text
        self.sections = doc.TextSections
        self.indexes = defaultdict(lambda: defaultdict(set))
        self.indexdialog()

    def indexdialog(self):
        try:
            # need this for inserting index into table
            table = self.viewcursor.TextTable
            if table:
                self.viewcursor.gotoRange(table.Anchor, False)
            self.doc.lockControllers()
            dialogprovider = self.smgr.createInstanceWithArguments("com.sun.star.awt.DialogProvider2", (self.doc,))
            uri = "$(user)/Scripts/python/multi_lex_index.xml"
            dialog_url = _substitute_variables(uri, self.smgr)
            dialog = dialogprovider.createDialogWithHandler(dialog_url, self)
            dialog.setTitle(self.TITLE)
            for name, label in zip(('lblCHOICE', 'btnOK', 'btnDELETE', 'btnCANCEL', 'lblALL'),
                                (self.LBL_CHOICE, self.LBL_INSERT, self.LBL_DELETE, self.LBL_CANCEL, self.LBL_ALL)):
                dialog.Model.getByName(name).Label = label
            self.getindexes(self.text)
            listbox = dialog.getControl("index")
            try:
                indexnames = sorted(self.indexes.keys(), key=locale.strxfrm)
            except:
                indexnames = sorted(self.indexes.keys())
            listbox.addItems(tuple(indexnames), 0)
            if indexnames:
                listbox.selectItemPos(0, True)
            dialog.execute()
        finally:
            self.doc.unlockControllers()

    def createallindexes(self, indexnames):
        try:
            for indexname in indexnames:
                if not self.sections.hasByName(indexname + self.CONTENT_SUFFIX):
                    self.createindex(indexname)
        except:
            traceback.print_exc()

    def createindex(self, indexname, withheader=True):
        if not indexname:
            return
        try:
            self.undo.enterUndoContext(self.MSG_INSERTED.format(indexname))
            if withheader:
                headersection = self.insertsection(indexname+self.HEADER_SUFFIX)
                headersection.Anchor.setString(self.capitalize(indexname))
                headersection.Anchor.ParaStyleName = "User Index Heading"
            contentsection = self.insertsection(indexname+self.CONTENT_SUFFIX)
            self.insertlines(contentsection, indexname)
        except:
            traceback.print_exc()
        finally:
            self.undo.leaveUndoContext()

    def insertsection(self, name, pos=None):
        section = self.doc.createInstance("com.sun.star.text.TextSection")
        section.setName(name)
        self.vctext.insertTextContent(self.viewcursor, section, True)
        return section

    def insertlines(self, section, indexname):
        def strxfrmy(seq):
            return [locale.strxfrm(x) for x in seq]

        def buildlines(entries_, level=1):
            for g, items in groupby(entries_, itemgetter(*range(level))):
                if level == 1:
                    g = (g,)
                if g in entries:
                    insertline(g, level, withpages=True)
                else:
                    insertline(g, level)
                    buildlines(items, level+1)

        def insertline(entry, level, withpages=False):
            if withpages:
                pages = ', '.join(self.getpagerange(index[entry]))
                line = self.PATTERN.format(entry[-1], pages)
            else:   #keys
                line = entry[level-1]
            section.Anchor.End.ParaStyleName = self.STYLES[level]
            T.insertString(section.Anchor.End, line, False)
            T.insertControlCharacter(section.Anchor.End, 5, False)
        index = self.indexes[indexname]
        try:
            entries = sorted(index.keys(), key=strxfrmy)
        except:
            entries = sorted(index.keys())
        T = section.Anchor.Text
        buildlines(entries)

    def updateallindexes(self, indexnames):
        try:
            for indexname in indexnames:
                if self.sections.hasByName(indexname+self.CONTENT_SUFFIX):
                    self.updateindex(indexname)
        except:
            traceback.print_exc()

    def updateindex(self, indexname):
        if not indexname:
            return
        try:
            self.undo.enterUndoContext(self.MSG_UPDATE.format(indexname))
            contentsection = self.sections.getByName(indexname+self.CONTENT_SUFFIX)
            self.viewcursor.gotoRange(contentsection.Anchor, False)
            contentsection.Anchor.setString('')
            self.insertlines(contentsection, indexname)
        except:
            traceback.print_exc()
        finally:
            self.undo.leaveUndoContext()

    def deleteallindexes(self, indexnames):
        try:
            for indexname in indexnames:
                if self.sections.hasByName(indexname+self.CONTENT_SUFFIX):
                    self.deleteindex(indexname)
        except:
            traceback.print_exc()

    def deleteindex(self, indexname):
        try:
            self.undo.enterUndoContext(self.MSG_CANCELLED.format(indexname))
            for name in (indexname+self.HEADER_SUFFIX, indexname+self.CONTENT_SUFFIX,):
                try:
                    section = self.sections.getByName(name)
                    anchor = section.Anchor
                    self.viewcursor.gotoRange(anchor.Start, False)
                    anchor.setString("")
                    section.dispose() 
                    c = anchor.Text.createTextCursor()
                    if not c.TextTable:
                        anchor.createEnumeration().nextElement().dispose()
                except NoSuchElementException:
                    pass
        finally:
            self.undo.leaveUndoContext()

    def getindexes(self, text):
        viewcursor = self.viewcursor
        paraenum = text.createEnumeration()
        offset = 0
        while paraenum.hasMoreElements():
            para = paraenum.nextElement()
            if para.PageNumberOffset:
                self.viewcursor.gotoRange(para.Start, False)
                offset = para.PageNumberOffset - self.viewcursor.Page
            try:
                portionenum = para.createEnumeration()
                while portionenum.hasMoreElements():
                    portion = portionenum.nextElement()
                    try:
                        mark = portion.DocumentIndexMark
                        if mark:
                            index = self.indexes[mark.UserIndexName]
                            if mark.MarkEntry:
                                entries = tuple(self.capitalize(x.strip()) for x in mark.MarkEntry.split(self.SEP))
                                viewcursor.gotoRange(portion, False)
                                index[entries].add(viewcursor.Page+offset)
                            else:
                                if portion.IsStart:
                                    viewcursor.gotoRange(portion.Start, False)
                                else:
                                    viewcursor.gotoRange(portion.Start, True)
                                    index[(self.capitalize(viewcursor.String),)].add(viewcursor.Page+offset)
                    except AttributeError:
                        pass
            except AttributeError:
                pass

        # check sub texts
        if text == self.text:
            frames = self.doc.TextFrames
            for n in range(frames.Count):
                frame = frames.getByIndex(n)
                self.getindexes(frame)
            tables = self.doc.TextTables
            for n in range(tables.Count):
                table = tables.getByIndex(n)
                for cellname in table.CellNames:
                    cell = table.getCellByName(cellname)
                    self.getindexes(cell)

        viewcursor.gotoRange(self.currentpos, False)

    def mergepages(self, indexname):
        '''obsolete'''
        indexmarks = self.indexes[indexname]
        for entry in indexmarks:
            pages = sorted(indexmarks[entry])
            pages = ', '.join(self.getpagerange(pages))
            indexmarks[entry] = "{} \t {}".format(entry[-1], pages)

    def getpagerange(self, pages):
        data = []
        for k, g in groupby(enumerate(pages), lambda x: x[0]-x[1]):
            g = list(g)
            d = "{}".format(sorted(g)[0][1])
            if len(g) > 1:
                d += "-{}".format(list(g)[-1][1])
            data.append(d)
        return data

    def callHandlerMethod(self, dialog, event, method):
        try:
            if method == 'choice':
                choice = event.Source.SelectedItem + self.CONTENT_SUFFIX
                btnOK = dialog.getControl("btnOK")
                btnOK.setEnable(True)
                dialog.getControl("lblALL").setEnable(True)
                exists = isinstance(self.sections.ElementNames, tuple) and choice in self.sections.ElementNames
                if exists:
                    btnOK.setLabel(self.LBL_UPDATE)
                    dialog.getControl("btnDELETE").setEnable(True)
                else:
                    btnOK.setLabel(self.LBL_INSERT)
                    dialog.getControl("btnDELETE").setEnable(False)
            elif method == "OK":
                choice = dialog.getControl("index").SelectedItem
                if event.Source.Model.Label == self.LBL_INSERT:
                    if dialog.getControl('lblALL').State==1:
                        self.createallindexes(dialog.getControl("index").Items)
                    else:
                        self.createindex(choice)
                elif event.Source.Model.Label == self.LBL_UPDATE:
                    if dialog.getControl('lblALL').State==1:
                        self.updateallindexes(dialog.getControl("index").Items)
                    else:
                        self.updateindex(choice)
                dialog.endExecute()
            elif method == "delete":
                if dialog.getControl('lblALL').State==1:
                    self.deleteallindexes(dialog.getControl("index").Items)
                else:
                    choice = dialog.getControl("index").SelectedItem
                    self.deleteindex(choice)
                dialog.endExecute()
            else:
                return False
        except:
            traceback.print_exc()

    def getSupportedMethodNames(self):
        return 'choice', 'OK', 'delete',

    def capitalize(self, s):
        return s[0].upper() + s[1:]

    def loadlocale(self, loc):
        for k, v in self.LOCALES.items():
            setattr(self, k, v[loc])

    def availablelocales(self):
        return self.LOCALES['TITLE'].keys()


def insertindex(event=None):
    ctx = XSCRIPTCONTEXT.getComponentContext()
    doc = XSCRIPTCONTEXT.getDocument()
    loc = _substitute_variables("$(vlang)", ctx.ServiceManager, False)
    if doc.supportsService("com.sun.star.text.GenericTextDocument"):
        indexcreator = IndexCreator(ctx, doc, loc)


def _substitute_variables(s, smgr, required=True):
    return smgr.createInstance("com.sun.star.util.PathSubstitution").substituteVariables(s, required)


g_exportedScripts = insertindex,
